==> space-element-0.0.4.tgz: package/src/components/space-element.js <==
const template = document.createElement("template");
template.innerHTML = `
	<style>
		/* box sizing */
		:host {
			box-sizing: border-box;
		}
		*, :host *:before, :host *:after {
			box-sizing: inherit;
		}
		/* variables */
		:host {
			--fs: 16px;
			--size: 1.3em;
			--font-family: Helvetica Neue, serif, sans-serif;
		}
		@media (min-width: 40rem) {
			:host {
				--fs: 20px;
			}
		}
		@media (prefers-color-scheme: dark) {
			:host {
				--c-link: aqua;
				--c-fg: white;
				--c-bg: black;
				--c-bg--modal: slategray;
				--c-bg--edit: white;
				--c-bg--button: transparent;
			}
		}
		@media (prefers-color-scheme: light) {
			:host {
				--c-bg: white;
				--c-fg: black;
				--c-link: magenta;
				--c-bg--modal: black;
				--c-bg--edit: gold;
			}
		}
		@media print {
			--fs: 17px;
		}
		/* layout */
		:host {
			display: flex;
			flex-direction: column;
			flex-grow: 1;
			font-size: var(--fs);
			font-family: var(--font-family);
			color: var(--c-fg);
			background-color: var(--c-bg);
			transition: background-color ease-in-out 200ms;
		}
		:host([is-editing="true"]) {
			background-color: var(--c-bg--edit);
		}
		@media print {
			:host {
				background-color: transparent;
			}
			slot[name="site"] {
				order: 1;
				page-break-after: always;
			}
			slot[name="settings"] {
				max-height: 100vh;
				overflow: scroll;
				order: 2;
				page-break-before: always;
				page-break-after: always;
			}
		}
		/* animations */
		/* initiate a double click */
		@keyframes initiateClick {
			0% {
				opacity: 1;
			}
			50% {
				opacity: 0;
			}
			55% {
				opacity: 1;
			}
			100% {
				opacity: 1;
			}
		}
		/* open modal */
		@keyframes openModal {
			0% {opacity: 0;}
			100% {opacity: 1;}
		}

		/* links */
		a,
		a:active {
			color: var(--c-link);
		}

		/*
		SLOT: SITE;
		*/
		slot[name="site"] {
			flex-grow: 1;
		}
		:host([is-editing="true"]) slot[name="site"] {
			display: none;
		}
		:host(:not([is-editing="true"])) slot[name="site"]{
			display: flex;
			flex-direction: column;
		}
		/* contains user generated styles */
		space-site {
			padding: var(--size);
			flex-grow: 1;
			outline-offset: calc(var(--size) / 3 * -1);
		}
		space-site:not([value]) {
			display: flex;
			flex-direction: column;
			justify-content: center;
			align-items: center;
			color: var(--c-link);
			overflow: hidden;
		}
		space-site:not([value])::before {
			content: '∅';
			font-size: 3rem;
			cursor: text;
			animation-name: initiateClick;
			animation-duration: 10s;
			animation-iteration-count: infinite;
		}

		/* user generated content */
		space-site h1,
		space-site h2,
		space-site h3,
		space-site h4,
		space-site h5,
		space-site h6,
		space-site p,
		space-site ul,
		space-site ol {
			margin-top: 0;
		}
		space-site a {
			word-break: break-word;
			white-space: break-spaces;
		}
		space-site figure {
			margin: 0;
		}
		space-site img {
			display: block;
			max-width: 100%;
			max-width: 100%;
			width: auto;
			height: 100%;
			max-height: 100vh;
		}
		space-site button {
			padding: calc(var(--size) / 2);
			cursor: pointer;
		}
		space-site button[dir="after"] {
			margin-bottom: var(--size);
		}
		/* a summary */
		space-site summary {
			font-size: 1.4em;
			cursor: pointer;
			display: flex;
			flex-direction: row;
			align-items: center;
			font-weight: bold;
			padding: calc(var(--size) / 2) 0;
		}
		space-site summary::before {
			color: var(--c-bg);
			content: '▶';
			float: left;
			margin-right: calc(var(--size) / 2);
			font-size: 1em;
			line-height: 1;
			transform: rotate(0);
			transition: transform 200ms ease-in-out, color 333ms ease-in-out;
		}
		space-site details[open] summary::before {
			transform: rotate(90deg);
			color: var(--c-fg);
		}
		space-site summary:hover::before {
			color: var(--c-link);
		}

		/* with the matrix element */
		space-site matrix-room-element {}
		space-site matrix-room-state {
			padding: var(--size);
			display: flex;
			flex-direction: column;
		}
		space-site matrix-event {
			display: block;
			margin-bottom: var(--size);
		}
		space-site matrix-event-track,
		space-site matrix-event-image {
			display: flex;
			flex-direction: column;
		}
		space-site matrix-event-image {
			flex-grow: 1;
			display: flex;
			flex-direction: column;
			align-items: center;
		}
		space-site matrix-event-image figure {
			display: flex;
			flex-direction: column;
			align-items: center;
		}
		space-site matrix-event-image figcaption {
			word-break: break-word;
		}
		space-site matrix-room-context {
			display: flex;
			flex-direction: column;
		}

		/*
		SLOT: SETTINGS;
		*/
		slot[name="settings"] {
			display: flex;
			flex-direction: column;
		}
		@media screen {
			slot[name="settings"] {
				justify-content: flex-end;
				position: sticky;
				bottom: 0;
			}
		}
		@media print {
			slot[name="settings"] {
				align-items: flex-end;
			}
		}
		/* side panel details */
		slot[name="settings"] details {}
		slot[name="settings"] details[open] {
			margin: calc(var(--size) / 4);
			background-color: var(--c-bg--modal);
			display: flex;
			flex-direction: column;
			justify-content: center;
			align-items: center;
			border-top-left-radius: 0.3em;
			border-top-right-radius: 0.3em;
		}
		slot[name="settings"] details[open] {
			animation-name: openModal;
			animation-duration: 0.4s;
		}

		/* summary */
		slot[name="settings"] summary {
			text-align: center;
			cursor: pointer;
			list-style: none;
			line-height: 1;
			border-top-left-radius: 0.3em;
			border-top-right-radius: 0.3em;
			overflow: hidden;
		}
		slot[name="settings"] summary:hover {
			color: var(--c-link);
		}
		slot[name="settings"] details[open] summary {
			position: sticky;
			top: 0;
			z-index: 1;
		}
		slot[name="settings"] summary::before {
			content: '↬';
			padding: calc(var(--size) / 5);
			display: block;
			font-size: 2em;
			transform: scale(1.2);
			transition: transform 333ms ease-in-out;
		}
		slot[name="settings"] details[open] summary::before {
			content: '×';
		}
		slot[name="settings"] summary:hover::before {
			transform: rotate(-33deg) scale(1.4);
		}
		slot[name="settings"] details[open] summary:hover::before {
			transform: scale(0.7) rotate(0);
		}
		@media print {
			slot[name="settings"] summary {
				display: none;
			}
		}
		/* buttons */
		slot[name="settings"] button {
			transition: background-color ease-in-out 150ms, border-color ease-in-out 180ms;
			/* margin: calc(var(--size) / 5); */
			padding: calc(var(--size) / 2);
			cursor: pointer;
			border-radius: 0;
			background-color: var(--c-bg--button);
			border: 0.2em solid var(--c-fg);
			color: var(--c-fg);
			font-size: 1em;
			border-radius: 0.2em;
		}
		slot[name="settings"] button:hover {
			border-color: var(--c-link);
		}
		/* menu */
		slot[name="settings"] space-element-menu {}
		slot[name="settings"] details[open] space-element-menu {
			position: sticky;
			bottom: 0;
			z-index: 1;
			background-color: var(--c-bg--modal);
		}
		slot[name="settings"] space-element-menu menu {
			text-align: right;
			padding: calc(var(--size) / 3);
			margin: 0;
			display: flex;
			justify-content: space-between;
			align-items: center;
		}
		slot[name="settings"] space-element-menu li {
			list-style: none;
			margin-top: 0;
		}
		slot[name="settings"] space-element-menu a {
			text-decoration: none;
			color: var(--c-fg);
		}
		slot[name="settings"] space-element-menu a:hover {
			color: var(--c-link);
		}
		@media print {
			slot[name="settings"] space-element-menu {
				display: none;
			}
		}
		/* special styles for the "edit" button, when "user is editing" */
		:host([is-editing="true"]) slot[name="settings"] button[name="edit"],
		:host([is-editing="true"]) slot[name="settings"] summary::before {
			color: var(--c-bg);
		}
		:host([is-editing="true"]) slot[name="settings"] button[name="edit"]:hover {
			border-color: var(--c-fg);
		}
		:host(:not([is-editing="true"])) slot[name="settings"] button[name="edit"]:hover {
			border-color: var(--c-bg);
		}

		/* qr code */
		slot[name="settings"] space-qrcode {
			display: flex;
			flex-wrap: nowrap;
			justify-content: center;
			align-items: center;
			flex-grow: 1;
		}
		slot[name="settings"] space-qrcode[is-active="true"] {
			position: fixed;
			top: 0;
			right: 0;
			bottom: 0;
			left: 0;
			background-color: var(--c-bg--modal);
			z-index: 2;
			border: 0.2em solid var(--c-link);
		}
		slot[name="settings"] qr-code {
			display: flex;
			max-height: 10vh;
			max-width: 10vh;
			transition: max-width 400ms ease-in-out;
			padding: calc(var(--size)/4);
			cursor: pointer;
			transform: scale(1);
			transition: transform 200ms ease-in-out;
		}
		slot[name="settings"] qr-code:hover {
			transform: scale(1.1);
		}
		slot[name="settings"] space-qrcode[is-active="true"] qr-code:hover {
			transition: none;
			transform: scale(1);
		}
		slot[name="settings"] space-qrcode[is-active="true"] qr-code {
			max-width: none;
			max-height: none;
			width: 100%;
			height: 100%;
		}

		slot[name="settings"] space-qrcode button[is-active] {
			background-color: var(--c-link);
		}
		slot[name="settings"] space-qrcode[is-active="true"] button {
			display: none;
		}
		@media print {
			slot[name="settings"] space-qrcode qr-code {
				max-width: 50vw;
				max-height: 50vh;
			}
			slot[name="settings"] space-qrcode button {
				display: none;
			}
		}

		/*
		SLOT: EDIT;
		*/
		:host([is-editing="true"]) slot[name="edit"] {
			/* display: inherit; */
		}
		:host(:not([is-editing="true"])) slot[name="edit"]{
			display: none;
		}
		space-edit {
			flex-grow: 1;
			display: flex;
			padding: calc(var(--size) / 3);
			width: 100%;
			overflow: visible;
		}
		space-edit textarea {
			resize: none;
			width: 100%;
			border: none;
			font-size: var(--fs);
			padding: var(--size);
			color: var(--c-fg);
			background-color: var(--c-bg);
			outline: 0.2em solid transparent;
			border-radius: 0.2em;
		}
	</style>
	<slot name="site">
		<space-site tabindex="0"></space-site>
	</slot>
	<slot name="edit">
		<space-edit></space-edit>
	</slot>
	<slot name="settings">
		<details>
			<summary title="Menu"></summary>
			<space-element-menu>
				<menu>
					<li>
						<space-qrcode></space-qrcode>
					</li>
					<li><a target="_blank" href="https://gitlab.com/sctlib/goog-space" title="About this website?">About</a></li>
					<li><button type="button" name="edit" title="Edit this space's content">Edit</button></li>
				</menu>
			</space-element-menu>
		</details>
	</slot>
`;

/**
 * Unicode to ASCII (encode data to Base64)
 * @param {string} data
 * @return {string}
	 https://base64.guru/developers/javascript/examples/unicode-strings
	 It preserves all unicodes characters (including emojis) when in a URL parameter
 */
function decode(data) {
	return decodeURIComponent(escape(atob(data)));
}
function encode(data) {
	return btoa(unescape(encodeURIComponent(data)));
}

export default class SpaceElement extends HTMLElement {
	static get observedAttributes() {
		return ["type", "is-editing", "space-title", "space-description"];
	}
	get type() {
		return this.getAttribute("type") || "default";
	}
	get isEditing() {
		return this.getAttribute("is-editing") === "true";
	}
	set isEditing(bool) {
		this.setAttribute("is-editing", bool);
	}
	get siteTitle() {
		return this.getAttribute("space-title") || "goog.space";
	}
	get siteDescription() {
		return this.getAttribute("space-description") || "spaces for all";
	}
	constructor() {
		super();
		this.setAttribute("tabindex", 0);
		this.$root = this.attachShadow({ mode: "open" });
		this.$root.appendChild(template.content.cloneNode(true));
		this.$root.addEventListener("keydown", this.handleKeyboard.bind(this));
		this.$modal = this.$root.querySelector("details");
		this.$site = this.$root.querySelector('slot[name="site"] space-site');
		// this.$site.addEventListener('dblclick', this.onSiteDoubleClick.bind(this))
		this.$site.setAttribute("space-title", this.siteTitle);
		this.$edit = this.$root.querySelector('slot[name="edit"] space-edit');
		this.$edit.addEventListener("input", this.onInput.bind(this));
		this.$edit.addEventListener("dblclick", this.onSiteDoubleClick.bind(this));
		this.$qrcode = this.$root.querySelector(
			'slot[name="settings"] space-qrcode'
		);
		const $editBtn = this.$root.querySelector(
			'slot[name="settings"] button[name="edit"]'
		);
		$editBtn.addEventListener("click", this.onEdit.bind(this));
	}
	connectedCallback() {
		this.data = this.readDataURL();
		if (this.data) {
			this.isEditing = false;
		} else if (!this.data) {
			this.isEditing = true;
		}
		this.renderMeta();
		this.render();
		this.focus();
	}

	readDataURL() {
		const { search, hash } = window.location;
		const hashParams = new URLSearchParams(hash.slice(1));
		const hashData = hashParams.get("data");
		const hashInput = hashParams.get("input");

		let data;
		if (hashData) {
			data = JSON.parse(decode(hashData));
		} else if (hashInput) {
			data = {
				type: "text",
				value: hashInput,
			};
		}

		if (!data) {
			const searchParams = new URLSearchParams(search);
			const searchData = searchParams.get("data");
			const searchInput = searchParams.get("input");
			if (searchData) {
				data = JSON.parse(decode(searchData));
			} else if (searchInput) {
				data = {
					type: "text",
					value: searchInput,
				};
			}
		}

		return data;
	}
	renderMeta() {
		const $head = document.querySelector("head");

		let $title = document.querySelector("title");
		if (!$title) {
			$title = document.createElement("title");
			$head.append($title);
		}
		$title.innerText = this.siteTitle;

		const $desc = document.createElement("meta");
		$desc.setAttribute("name", "description");
		$desc.setAttribute("content", this.siteDescription);
		$head.append($desc);
	}
	render() {
		if (this.data) {
			this.renderData(this.data.value);
		} else if (!this.data) {
			this.renderData("");
		}
		this.renderQrcode(window.location);
	}
	renderData(value) {
		if (value) {
			this.$site.setAttribute("value", JSON.stringify(value));
			this.$edit.setAttribute("value", JSON.stringify(value));
		} else {
			this.$site.removeAttribute("value");
			this.$edit.removeAttribute("value");
		}
	}
	renderQrcode(value) {
		value && this.$qrcode.setAttribute("value", value);
	}
	onInput(event) {
		const data = {
			type: this.type,
			value: event.target.value,
		};

		/* update URL query param */
		const { search } = window.location;
		const params = new URLSearchParams(search);
		if (data.value) {
			params.set("data", encode(JSON.stringify(data)));
			history.replaceState(null, null, "#" + params.toString());
		} else {
			params.delete("data");
			history.replaceState(null, null, "#" + params.toString());
			/* history.replaceState(window.location.origin + window.location.pathname, null, params.toString()) */
		}

		/* update qrcode */
		this.renderQrcode(window.location);

		/* update value of edit area, and rendered site area */
		this.renderData(event.target.value);
	}
	onEdit(event) {
		this.isEditing = !this.isEditing;
		if (this.isEditing) {
			this.$edit.querySelector("textarea").focus();
		} else {
			this.$site.focus();
		}
	}
	onSiteDoubleClick(event) {
		this.onEdit();
	}
	handleKeyboard(event) {
		/* on escape, close modal, or cancel editting */
		if (event.key === "Escape") {
			if (this.$modal.open) {
				this.$modal.open = false;
			} else {
				if (this.isEditing) {
					this.onEdit();
				}
			}
		}
		if (event.key === "Enter" && event.ctrlKey) {
			this.onEdit();
		}
	}
}

==> space-element-0.0.6.tgz: package/src/components/space-element.js <==
import spaceElementStyles from "../styles/space-element.css?inline";
import * as Y from "yjs";
import debounce from "lodash.debounce";

const template = document.createElement("template");
template.innerHTML = `
	<style>
		${spaceElementStyles}
	</style>
	<slot name="site">
		<space-site tabindex="0"></space-site>
	</slot>
	<slot name="edit">
		<space-edit></space-edit>
	</slot>
	<slot name="settings">
		<details>
			<summary title="Menu"></summary>
			<space-element-menu>
				<menu>
					<li>
						<space-qrcode></space-qrcode>
					</li>
					<li><a target="_blank" href="https://gitlab.com/sctlib/goog-space" title="About this website?">About</a></li>
					<li><button type="button" name="edit" title="Edit this space's content">Edit</button></li>
					<li><rtc-user></rtc-user></li>
				</menu>
			</space-element-menu>
		</details>
	</slot>
`;

/**
 * Unicode to ASCII (encode data to Base64)
 * @param {string} data
 * @return {string}
	 https://base64.guru/developers/javascript/examples/unicode-strings
	 It preserves all unicodes characters (including emojis) when in a URL parameter
 */
function decode(data) {
	return decodeURIComponent(escape(atob(data)));
}
function encode(data) {
	return btoa(unescape(encodeURIComponent(data)));
}

export default class SpaceElement extends HTMLElement {
	static get observedAttributes() {
		return ["type", "is-editing", "space-title", "space-description"];
	}
	get type() {
		return this.getAttribute("type") || "default";
	}
	get isEditing() {
		return this.getAttribute("is-editing") === "true";
	}
	set isEditing(bool) {
		this.setAttribute("is-editing", bool);
	}
	get siteTitle() {
		return this.getAttribute("space-title") || "goog.space";
	}
	get siteDescription() {
		return this.getAttribute("space-description") || "spaces for all";
	}
	get debounceTime() {
		return 333;
	}
	/* apply local changes and send updates */
	updateYDoc(data = "") {
		this.yDoc.transact(() => {
			this.yData.delete(0, this.yData.toString().length);
			this.yData.insert(0, data);
		});
	}
	updateRtc() {
		const yUpdate = Y.encodeStateAsUpdate(this.yDoc);
		if (this.$rtcUser.dataChannel) {
			this.$rtcUser.send(yUpdate);
		}
	}
	updateUrl(data) {
		const { search } = window.location;
		const params = new URLSearchParams(search);
		if (data.value) {
			params.set("data", encode(JSON.stringify(data)));
			history.replaceState(null, this.siteTitle, "#" + params.toString());
		} else {
			params.delete("data");
			if (params.size) {
				history.replaceState(null, this.siteTitle, "#" + params.toString());
			} else {
				history.pushState(
					"",
					this.siteTitle,
					window.location.pathname + window.location.search
				);
			}
		}
	}
	readDataURL() {
		const { search, hash } = window.location;
		const hashParams = new URLSearchParams(hash.slice(1));
		const hashData = hashParams.get("data");
		const hashInput = hashParams.get("input");

		let data;
		if (hashData) {
			data = JSON.parse(decode(hashData));
		} else if (hashInput) {
			data = {
				type: "text",
				value: hashInput,
			};
		}

		if (!data) {
			const searchParams = new URLSearchParams(search);
			const searchData = searchParams.get("data");
			const searchInput = searchParams.get("input");
			if (searchData) {
				data = JSON.parse(decode(searchData));
			} else if (searchInput) {
				data = {
					type: "text",
					value: searchInput,
				};
			}
		}
		return data;
	}

	constructor() {
		super();
		this.setAttribute("tabindex", 0);
		this.data = {};
		this.$root = this.attachShadow({ mode: "open" });
		this.$root.appendChild(template.content.cloneNode(true));
		this.$root.addEventListener("keydown", this.onKeyboard.bind(this));
		this.$modal = this.$root.querySelector("details");
		this.$site = this.$root.querySelector('slot[name="site"] space-site');
		// this.$site.addEventListener('dblclick', this.onSiteDoubleClick.bind(this))
		this.$site.setAttribute("space-title", this.siteTitle);
		this.$edit = this.$root.querySelector('slot[name="edit"] space-edit');
		this.$edit.addEventListener("input", this.onInput.bind(this));
		this.$edit.addEventListener("dblclick", this.onSiteDoubleClick.bind(this));
		this.$qrcode = this.$root.querySelector(
			'slot[name="settings"] space-qrcode'
		);
		const $editBtn = this.$root.querySelector(
			'slot[name="settings"] button[name="edit"]'
		);
		$editBtn.addEventListener("click", this.onEdit.bind(this));

		this.$rtcUser = this.$root.querySelector('slot[name="settings"] rtc-user');
		this.$rtcUser.addEventListener(
			"dataChannel",
			this.onDataChannel.bind(this)
		);
		this.dataChannels = [];
		this.$rtcUser.addEventListener(
			"channelMessage",
			this.onChannelMessage.bind(this)
		);
		// have to debounce setting the URL params
		this.debouncedUpdateUrl = debounce(this.updateUrl, this.debounceTime);
	}
	connectedCallback() {
		this.data = this.readDataURL();
		this.yDoc = new Y.Doc();
		this.yData = this.yDoc.getText("data", Y.XmlText);
		/* this.yDoc.on("update", this.onYDocUpdate.bind(this)); */
		this.yDoc.on("afterTransaction", this.onYDocUpdate.bind(this));
		if (this.data) {
			this.isEditing = false;
			this.yData.insert(0, this.data.value);
		} else if (!this.data) {
			this.isEditing = true;
		}
		this.renderMeta();
		this.render();
		this.focus();
	}
	renderMeta() {
		const $head = document.querySelector("head");

		let $title = document.querySelector("title");
		if (!$title) {
			$title = document.createElement("title");
			$head.append($title);
		}
		$title.innerText = this.siteTitle;

		const $desc = document.createElement("meta");
		$desc.setAttribute("name", "description");
		$desc.setAttribute("content", this.siteDescription);
		$head.append($desc);
	}
	render() {
		if (this.data) {
			this.renderData(this.data.value);
		} else if (!this.data) {
			this.renderData("");
		}
		this.renderQrcode();
	}
	renderData(value) {
		if (value) {
			this.$site.setAttribute("value", JSON.stringify(value));
			this.$edit.setAttribute("value", JSON.stringify(value));
		} else {
			this.$site.removeAttribute("value");
			this.$edit.removeAttribute("value");
		}
	}
	renderQrcode(value = window.location) {
		value && this.$qrcode.setAttribute("value", value);
	}

	onYDocUpdate() {
		const data = {
			value: this.yData.toString(),
		};
		this.debouncedUpdateUrl(data);
		this.renderQrcode();
		this.renderData(data.value);
	}
	onInput(event) {
		this.updateYDoc(event.target.value);
		this.updateRtc();
	}
	onPeerInput(yString) {}
	onEdit(event) {
		this.isEditing = !this.isEditing;
		if (this.isEditing) {
			this.$edit.querySelector("textarea").focus();
		} else {
			this.$site.focus();
		}
	}
	onSiteDoubleClick(event) {
		this.onEdit();
	}
	onKeyboard(event) {
		/* on escape, close modal, or cancel editting */
		if (event.key === "Escape") {
			if (this.$modal.open) {
				this.$modal.open = false;
			} else {
				if (this.isEditing) {
					this.onEdit();
				}
			}
		}
		if (event.key === "Enter" && event.ctrlKey) {
			this.onEdit();
		}
	}
	/* send data to peer (both do the same, to share their initial state) */
	onDataChannel(event) {
		this.updateRtc();
	}
	/* update our yDoc data with the peer data, syncs all states on changes */
	onChannelMessage({ detail }) {
		let reader = new FileReader();
		reader.onloadend = () => {
			const update = new Uint8Array(reader.result);
			Y.applyUpdate(this.yDoc, update);
		};
		reader.readAsArrayBuffer(detail);
	}
}
